using System;
using Microsoft.SPOT;

namespace Gadgeteer.Modules.Community.XBee
{
    /// <summary>
    /// Base class for all tests
    /// </summary>
    public abstract class BaseTest
    {
        /// <summary>
        /// Sets property defaults
        /// </summary>
        public BaseTest()
        {
            this.TimeoutMilliseconds = 10000;
            this.HaltTestRunOnFailure = false;
            this.Message = String.Empty;
            this.Passed = false;
        }

        /// <summary>
        /// if this test fails, halt further execution, this is useful for usability
        /// so that users will concentrate on this test failure and fix it.
        /// Further tests may rely on the results of this test
        /// </summary>
        public bool HaltTestRunOnFailure { get; set; }

        /// <summary>
        /// When your test passes, it may be useful to put a message here.
        /// It will be printed along with the test results.
        /// For example, an AT test issuing an SH command could put the SH value here, 
        /// Then the tests can be used for seeing lots of interesting information about your environment
        /// </summary>
        public string Message { get; set; }

        /// <summary>
        /// Test timeout in Milliseconds
        /// </summary>
        public int TimeoutMilliseconds { get; set; }

        /// <summary>
        /// Succeeded?
        /// </summary>
        public bool Passed { get; set; }

        /// <summary>
        /// If error occurs or assert fails
        /// </summary>
        public Exception Error { get; set; }

        /// <summary>
        /// The time it took to succeed or fail
        /// </summary>
        public long ElapsedMilliseconds { get; set; }

        /// <summary>
        /// test setup
        /// </summary>
        /// <param name="context">a TestContext</param>
        public virtual void Arrange(TestContext context)
        {

        }
        
        /// <summary>
        /// Runs the test actions
        /// </summary>
        /// <param name="context">a TestContext</param>
        public abstract void Act(TestContext context);

        /// <summary>
        /// throws XBeeClientTest exceptions or passes, this method will be retried until success or timeout.
        /// the exception should contain any tips on how to fix the issue
        /// </summary>
        /// <param name="context">a TestContext</param>
        public abstract void Assert(TestContext context);

        /// <summary>
        /// Usually for unsubscribing event handlers
        /// </summary>
        /// <param name="context">a TestContext</param>
        public virtual void Teardown(TestContext context)
        {

        }

        /// <summary>
        /// Formatted string of test
        /// </summary>
        /// <returns>Formatted string of test</returns>
        public override string ToString()
        {
            return "\n\nTestName: " + this.GetType().Name +
                   "\nSuccess: " + Passed +
                   "\nElapsedMilliseconds: " + ElapsedMilliseconds + 
                   "\nMessage: " + Message +
                   ((Error != null) ? "\nError: " + Error.ToString() : String.Empty);
        }
    }
}
