using System;
using Microsoft.SPOT;

namespace Gadgeteer.Modules.Community.XBee
{
    /// <summary>
    /// Base class for all AT tests
    /// </summary>
    public abstract class BaseAT_Test : BaseTest
    {
        ATResponseReceivedEventArgs eventArgs;

        /// <summary>
        /// Wire up the ATResponseReceived event handler
        /// </summary>
        /// <param name="context">TestContext</param>
        public override void Arrange(TestContext context)
        {
            context.XBee.ATResponseReceived += new ATResponseReceivedEventHandler(XBee_ATResponseReceived);
        }

        /// <summary>
        /// Gets the AT command, e.g. 'SH'
        /// </summary>
        /// <returns>derived classes AT Command</returns>
        protected abstract string GetATCommand();

        /// <summary>
        /// issue the SH command
        /// </summary>
        /// <param name="context">TestContext</param>
        public override void Act(TestContext context)
        {
            context.XBee.SendLocalATCommandAsync(GetATCommand());
        }

        void XBee_ATResponseReceived(object sender, ATResponseReceivedEventArgs e)
        {
            this.eventArgs = e;
        }

        /// <summary>
        /// If the test succeeds, the derived class will use the eventArgs to format the Message property
        /// </summary>
        /// <param name="eventArgs"></param>
        public abstract void Success(ATResponseReceivedEventArgs eventArgs);

        /// <summary>
        /// Validate that no error occured, CommandStatus is 0x00, and record the SH value in Message
        /// </summary>
        /// <param name="context">TestContext</param>
        public override void Assert(TestContext context)
        {
            if (null == eventArgs)
            {
                throw new XBeeAssertFailed("Did not receive an ATResponseReceived event");
            }
            else if (eventArgs.Error != null)
            {
                throw new XBeeAssertFailed(eventArgs.Error.ToString());
            }
            else if (eventArgs.Result.CommandStatus != 0x00)
            {
                throw new XBeeAssertFailed("CommandStatus not equal to OK, was " + HexUtil.ToHexString((byte)eventArgs.Result.CommandStatus));
            }
            else
            {
                Success(eventArgs);
            }

        }

        /// <summary>
        /// remove event handler
        /// </summary>
        /// <param name="context">TestContext</param>
        public override void Teardown(TestContext context)
        {
            context.XBee.ATResponseReceived -= XBee_ATResponseReceived;
        }
    }

    /// <summary>
    /// Issue AT command SH (Serial number High)
    /// </summary>
    public class SH_Test : BaseAT_Test
    {
        /// <summary>
        /// The AT Command for SerialNumber High
        /// </summary>
        /// <returns>'SH'</returns>
        protected override string GetATCommand()
        {
            return "SH";
        }

        /// <summary>
        /// Sets a message for the user
        /// </summary>
        /// <param name="eventArgs">ATResponseReceivedEventArgs</param>
        public override void Success(ATResponseReceivedEventArgs eventArgs)
        {
            Message = "SH value is " + HexUtil.ToHexString(eventArgs.Result.ReturnedBytes);
        }
    }
}
