using System;
using Microsoft.SPOT;
using Gadgeteer.Modules.Community.XBee;
using System.Collections;

namespace Gadgeteer.Modules.Community.XBee
{
    /// <summary>
    /// Send a small byte array to all peer nodes one at a time
    /// </summary>
    public class SmallUploadDataAsyncTest : BaseUploadDataAsyncTest
    {
        /// <summary>
        /// creates test and sets base class properties
        /// </summary>
        public SmallUploadDataAsyncTest()
            : base(new byte[]{0x01,0x02}, 1, false)
        {

        }
    }

    /// <summary>
    /// Issues UploadDataAsync for each Peer node
    /// </summary>
    public abstract class BaseUploadDataAsyncTest : BaseTest
    {
        ArrayList eventArgs;

        /// <summary>
        /// constructor to be called by all derived class constructors
        /// </summary>
        /// <param name="dataToSend">test data</param>
        /// <param name="numExpectedPackets">number of packets expected</param>
        /// <param name="isBroadcast">broadcast?</param>
        protected BaseUploadDataAsyncTest(byte[] dataToSend, int numExpectedPackets, bool isBroadcast)
        {
            this.DataToSend = dataToSend;
            this.NumExpectedPackets = numExpectedPackets;
            this.IsBroadast = isBroadcast;
        }

        /// <summary>
        /// The string the derived test will send
        /// </summary>
        public byte[] DataToSend { get; set; }

        /// <summary>
        /// Number of packets that should be created
        /// </summary>
        public int NumExpectedPackets { get; set; }

        /// <summary>
        /// Is this a broadcast message or should it be sent to individual nodes
        /// </summary>
        public bool IsBroadast { get; set; }

        /// <summary>
        /// Wire UploadStringCompleted handler
        /// </summary>
        /// <param name="context">TestContext</param>
        public override void Arrange(TestContext context)
        {
            context.XBee.TransmitStatusReceived += new TransmitStatusReceivedEventHandler(XBee_TransmitStatusReceived);
            eventArgs = new ArrayList();
        }

        void XBee_TransmitStatusReceived(object sender, TransmitStatusReceivedEventArgs e)
        {
            eventArgs.Add(e);
        }

        /// <summary>
        /// Issues UploadStringAsync for each Peer
        /// </summary>
        /// <param name="context">TestContext</param>
        public override void Act(TestContext context)
        {
            if (IsBroadast)
            {
                context.XBee.UploadDataAsync(DataToSend);
            }
            else
            {
                for (int i = 0; i < context.Peers.Length; i++)
                {
                    context.XBee.UploadDataAsync(context.Peers[i], DataToSend);
                }
            }
        }

        /// <summary>
        /// Validate that event handler called N times for N peers, and that the Error is null
        /// </summary>
        /// <param name="context">TestContext</param>
        public override void Assert(TestContext context)
        {
            int totalExpectedPackets = IsBroadast ? NumExpectedPackets : context.Peers.Length * NumExpectedPackets;

            int numReceivedOkPackets = 0;
            foreach (TransmitStatusReceivedEventArgs args in this.eventArgs)
            {
                if (args.Result.DeliveryStatus == 0x00)
                {
                    numReceivedOkPackets++;
                }
            }

            if (numReceivedOkPackets != totalExpectedPackets)
            {
                throw new XBeeAssertFailed("TransmitStatusReceived event handler only called " + numReceivedOkPackets + " times with DeliveryStatus OK, should have been called " + totalExpectedPackets + " times");
            }

            foreach (TransmitStatusReceivedEventArgs args in this.eventArgs)
            {
                if (args.Error != null)
                {
                    throw new XBeeAssertFailed(args.Error.ToString());
                }
                else if (args.Result.DeliveryStatus != 0x00)
                {
                    throw new XBeeAssertFailed("Deliver status was " + HexUtil.ToHexString((byte)args.Result.DeliveryStatus) + " but should be 0x00 (OK)");
                }

            }
        }

        /// <summary>
        /// unwire the event handler
        /// </summary>
        /// <param name="context">TestContext</param>
        public override void Teardown(TestContext context)
        {
            context.XBee.TransmitStatusReceived -= XBee_TransmitStatusReceived;
        }
    }
}
